package dev.kaderli.magsend.activity

import android.os.Bundle
import android.view.View
import android.widget.TextView
import androidx.core.content.ContextCompat
import dev.kaderli.magsend.Constants
import dev.kaderli.magsend.R
import dev.kaderli.magsend.Utility
import dev.kaderli.magsend.model.Sample
import dev.kaderli.magsend.model.Signal
import java.lang.Integer.min
import java.util.stream.Collectors
import kotlin.math.abs
import kotlin.math.pow

class ReceiveActivity : BaseSensorActivity() {
  private lateinit var receiveDescription: TextView
  private lateinit var preambleStatus: TextView
  private lateinit var crcStatus: TextView
  private lateinit var headerStatus: TextView
  private lateinit var receiveValue: TextView

  /**
   * List that contains all received samples from the sensor.
   */
  private var samples = ArrayList<Sample<Float>>()

  /**
   * A list that contains the samples as high or low signals.
   */
  private var signal = ArrayList<Sample<Signal>>()

  /**
   * Flag that determines if the preamble was already received.
   */
  private var preambleReceived = false

  private var payloadLength: Int? = null

  private var receivingComplete = false

  override fun onCreate(savedInstanceState: Bundle?) {
    super.onCreate(savedInstanceState)
    receiveDescription = findViewById(R.id.receiveDescription)
    preambleStatus = findViewById(R.id.preambleStatus)
    crcStatus = findViewById(R.id.crcStatus)
    headerStatus = findViewById(R.id.headerStatus)
    receiveValue = findViewById(R.id.receiveValue)
  }

  override fun getLayoutId(): Int {
    return R.layout.activity_receive
  }

  private fun getRangeOfData(): Float {
    val maxValue = samples.maxByOrNull { it.value } ?: Sample(0f)
    val minValue = samples.minByOrNull { it.value } ?: Sample(0f)

    return maxValue.value - minValue.value
  }

  private fun getThresholdForHighSignals(): Float {
    if (getRangeOfData() < Constants.SIGNAL_THRESHOLD) {
      return getMidRangeOfData() + Constants.SIGNAL_THRESHOLD
    }

    return getMidRangeOfData() + Constants.SIGNAL_THRESHOLD
  }

  private fun getThresholdForLowSignals(): Float {
    if (getRangeOfData() < Constants.SIGNAL_THRESHOLD) {
      return getMidRangeOfData() + Constants.SIGNAL_THRESHOLD
    }

    return getMidRangeOfData() - Constants.SIGNAL_THRESHOLD
  }

  private fun getMidRangeOfData(): Float {
    val maxValue = samples.maxByOrNull { it.value } ?: Sample(0f)
    val minValue = samples.minByOrNull { it.value } ?: Sample(0f)
    return (maxValue.value + minValue.value) / 2
  }

  private fun isHigh(value: Float): Boolean {
    return value > getThresholdForHighSignals()
  }

  private fun isLow(value: Float): Boolean {
    return value < getThresholdForLowSignals()
  }

  /**
   * Check if the given sample should count as a high or low signal. If the sample is inside
   * the threshold area it is not counted.
   */
  private fun getSignalForSample(sample: Sample<Float>): Sample<Signal>? {
    if (isHigh(sample.value)) {
      return Sample(Signal.High)
    }

    if (isLow(sample.value)) {
      return Sample(Signal.Low)
    }

    return null
  }

  /**
   * Attempts to detect the preamble in the already received signal. If it can find the preamble it returns true and
   * false otherwise.
   */
  private fun detectPreamble(): Boolean {
    val cleanedSignal = cleanSignal(signal)
    val compressedSignal = compressSignal(cleanedSignal)

    if (compressedSignal.size < 4) {
      return false
    }

    for (i in 0..compressedSignal.size - 4) {
      val firstValue = compressedSignal[i]
      val secondValue = compressedSignal[i + 1]
      val thirdValue = compressedSignal[i + 2]
      val fourthValue = compressedSignal[i + 3]

      if (firstValue.value != Signal.High || secondValue.value != Signal.Low || thirdValue.value != Signal.High) {
        continue
      }

      if (secondValue.timestamp - firstValue.timestamp > 2900 && thirdValue.timestamp - secondValue.timestamp > 2900) {
        return true
      }
    }

    return false
  }

  /**
   * This functions takes the signal and filters out samples that are only in a cluster of 3 or less of the same signal.
   * This way we can remove single jumps that occur in the signal.
   */
  private fun cleanSignal(signal: List<Sample<Signal>>): List<Sample<Signal>> {
    val cleanSignal = signal.filterIndexed { index, sample ->
      var elementCount = 1

      var previousIndex = index - 1
      while (previousIndex > 0 && signal[previousIndex].value == sample.value) {
        elementCount++
        previousIndex--
      }

      var nextIndex = index + 1
      while (nextIndex < signal.size && signal[nextIndex].value == sample.value) {
        elementCount++
        nextIndex++
      }

      elementCount >= Constants.SIGNAL_SAMPLE_THRESHOLD
    }



    return cleanSignal
  }

  private fun compressSignal(signal: List<Sample<Signal>>): List<Sample<Signal>> {
    // Only keep the first of consecutive same signals
    val compactSignal = ArrayList<Sample<Signal>>()
    signal.forEach { sample ->
      if (sample.value != compactSignal.lastOrNull()?.value) {
        compactSignal.add(sample)
      }
    }

    return compactSignal
  }

  private fun analyzePacketSignal(): List<Int> {
    val cleanedSignal = cleanSignal(signal)

    var maxSamplesInARow = 0
    var count = 0
    cleanedSignal.forEachIndexed { index, sample ->
      if (index == 0) {
        return@forEachIndexed
      }

      if (sample.value == cleanedSignal[index - 1].value) {
        count++
        if (count > maxSamplesInARow) {
          maxSamplesInARow = count
        }
      } else {
        count = 0
      }
    }

    val packetData = ArrayList<Int>()
    var lastSignal: Sample<Signal>? = null
    var isFirstSignal = true
    var samplesSinceLastSignal = 0
    cleanedSignal.forEach { sample ->
      if (lastSignal == null) {
        lastSignal = sample
      }

      if (sample.value != lastSignal!!.value) {
        if (samplesSinceLastSignal > (maxSamplesInARow * 0.75) || isFirstSignal) {
          if (sample.value == Signal.High) {
            packetData.add(0)
          } else {
            packetData.add(1)
          }
          isFirstSignal = false
          samplesSinceLastSignal = 0
        }
        lastSignal = sample
      }
      samplesSinceLastSignal++
    }

    return packetData
  }

  override fun sensorValueReceived(magneticFieldStrength: Float) {
    if (receivingComplete) {
      return;
    }

    // Add the current received sensor value to the samples
    samples.add(Sample(magneticFieldStrength))

    // Determine whether the sample is a high or low signal
    val currentSignal = getSignalForSample(samples.last()) ?: return
    signal.add(currentSignal)

    // Try to detect the preamble in the signal
    if (!preambleReceived) {
      // If we detect the preamble we change the status on the screen and clear the existing signal data.
      if (detectPreamble()) {
        preambleReceived = true
        preambleStatus.setText(R.string.preamble_status_detected)
        preambleStatus.setTextColor(ContextCompat.getColor(this, R.color.success))
        signal.clear()
        return
      }
    }

    // After the preamble is detected we attempt to decode the packet.
    if (preambleReceived) {
      // Get the decoded packet data
      val packet = analyzePacketSignal()

      // Get the payload length of in the data
      if (payloadLength == null && packet.size >= Constants.HEADER_LENGTH) {
        payloadLength = listToInteger(packet.take(Constants.HEADER_LENGTH))
        headerStatus.text = getString(R.string.payload_length, payloadLength)
      }

      if(payloadLength == null) {
        return
      }

      if (packet.size >= Constants.HEADER_LENGTH + Constants.BYTE_LENGTH) {
        val numberOfAvailablePayloadBytes = min((packet.size - Constants.HEADER_LENGTH) / 8, payloadLength!!)
        var payload = ""
        for (i in 1..numberOfAvailablePayloadBytes) {
          payload += listToInteger(
            packet.take(Constants.HEADER_LENGTH + (Constants.BYTE_LENGTH * i)).takeLast(Constants.BYTE_LENGTH)
          ).toChar().toString()
        }
        receiveValue.text = payload
      }

      if (packet.size >= Constants.HEADER_LENGTH + (Constants.BYTE_LENGTH * payloadLength!!) + Constants.CRC_LENGTH) {
        // CRC Check
        val receivedCrc = listToInteger(
          packet.take(Constants.HEADER_LENGTH + (Constants.BYTE_LENGTH * payloadLength!!) + Constants.CRC_LENGTH)
            .takeLast(Constants.CRC_LENGTH)
        )
        val calculatedCrc =
          Utility.crc8Autosar(receiveValue.text.codePoints().boxed().collect(Collectors.toList()))

        if (receivedCrc == calculatedCrc) {
          crcStatus.setText(R.string.crc_status_valid)
          crcStatus.setTextColor(ContextCompat.getColor(this, R.color.success))
        } else {
          crcStatus.setText(R.string.crc_status_invalid)
          crcStatus.setTextColor(ContextCompat.getColor(this, R.color.error))
        }

        // Mark receiving as complete
        receiveDescription.setText(R.string.receive_description_complete)
        receivingComplete = true
      }
    }
  }

  private fun listToInteger(values: List<Int>): Int {
    var value = 0

    for (bit in values) {
      value = (value shl 1) + bit
    }

    return value
  }


  fun restartReceiveProcess(view: View) {
    preambleReceived = false
    payloadLength = null
    receivingComplete = false

    receiveDescription.setText(R.string.receive_description)
    preambleStatus.setText(R.string.preamble_status_not_detected)
    preambleStatus.setTextColor(ContextCompat.getColor(this, R.color.error))
    headerStatus.text = ""
    crcStatus.text = ""
    receiveValue.text = ""

    samples.clear()
    signal.clear()
  }
}
